/* @(#)pymidas-layer1.cpp 1.0 (ESO-SAMPO) 03/09/03 12:02:40 */
//########################################################################
//
//   This file is part of PyMidas.
//
//   PyMidas is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   PyMidas is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with PyMidas; if not, write to the Free Software
//   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//
//###########################################################################
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
.COPYRIGHT  (c) 2005 European Southern Observatory
.LANGUAGE   C/C++
.IDENT      pymidas-layer1.c
.AUTHOR     T.Oittinen
.KEYWORDS    
.ENVIRON    UNIX
.PURPOSE    Layer between Python script and ESO-MIDAS
.VERSION    0.1   2005-Mar-09 : Creation, TO
------------------------------------------------------------------------*/
//#include <stdio.h> 
#include <stdarg.h>  
#include <signal.h> 
#include <unistd.h>
#include <pthread.h>
#include <stdlib.h>
 
#include <Python.h> 
#include <modsupport.h>

#include <midas_def.h>    
#include "pymidas-layer1.h"

#include <osyparms.h>
#include <monitdef.h>
#include <midfront.h>
#include <midback.h>  
#include <fsydef.h>

PyMODINIT_FUNC initPyMidas(void);
   
#define MAX_RETBUF          1000

void LayerProlog();

int giMidasPrologDone = 0;
int is_a_tty = 1;           /* Is this a terminal, (yes=1) set in prepa.c */

static int giInitDone = 0;
char gszMidasUnit[10]="";
char gszMyUnit[10]="";
char *gpszMidasUnit = gszMidasUnit;
char *gpszMyUnit    = gszMyUnit;
char gszRespFilename[100]; 
char szDummy[2] = " ";
char *gpResponseBuf = szDummy;
long glBufSize = 0L;
int  giCid = 0;
int  giCmdReq = 0;
int  giCmdResp = 0;
int  giCmdReqRemoved = -1;

//------------------------------------------------------------------------

ParamType aParams[] =
{
	{PAR_TRACE_LEVEL,"0","0",1},  		// 0=none, 1=all
	{PAR_TRACE_CONSOLE,"0","0",1},  	// 0=none, 1=to console
	{PAR_TRACE_FILE,"1","1",1},  		// 0=none, 1=to file
	{PAR_MIDAS_CMD_TIMEOUT,"5","5",1},  // default 5 sec
	{PAR_COMM_MODE,"0","0",1},  		// 0=files, 1=sockets
    {PAR_COMM_OUT,"0","0",1},           // 0=file,  1=file+terminal
	{"","","",0}
};

//------------------------------------------------------------------------
int ReadParamInt(char *pszParamName);
long ReadResponse( char **ppResp, int *piRet );
long ReadResponseFile( FILE *fp, char **ppResp );
int Initialize(char *pszMidasUnit, char *pszMyUnit, int iCid);

//------------------------------------------------------------------------
void intermail()
{
    return;
}

//------------------------------------------------------------------------
// RemoveResponseFile()
//
// Remove response files specified by param iCmd
//
void RemoveResponseFile( int iCmd )
{
    static char szFilename[100];
    
    //sprintf(szRespFilename, "pymidas%2.2s%1d.%04d", 
    //    gszMidasUnit, giCid, giCmdReq++);

    if ( iCmd < 0 ) return;
    
    sprintf(szFilename, "%s"FILENAME_FORMAT, // "%spymidas%2.2s%1d.%04d", 
        getenv("MID_WORK"), gszMidasUnit, giCid, iCmd);
    //sprintf(szFilename, "%spymidas%2.2s%1d.*", 
     //   getenv("MID_WORK"), gszMidasUnit, giCid);
    //sprintf(szFilename,"%s%s",getenv("MID_WORK"),szRespFilename);
    DBG((FL,"\t\tRemoving: %s\n", szFilename));
    remove(szFilename);
}//RemoveResponseFile()

//------------------------------------------------------------------------
// GetNewResponseFileNameReq()
//
// Build and returns response file name for request
//
char * GetNewResponseFileNameReq()
{
    static char szRespFilename[100];
	static char szFilename[100];
    int iParCommOut  = ReadParamInt(PAR_COMM_OUT);
    
    //
    // Remove old response files
    //
    DBG((FL,"==> GetNewResponseFileNameReq() - begin\n"));
    DBG((FL,"\t1: giCmdReq=%d, giCmdResp=%d, giCmdReqRemoved=%d\n", 
            giCmdReq, giCmdResp, giCmdReqRemoved));

    giCmdResp = giCmdReq;
     
    for ( ; giCmdReqRemoved < giCmdReq; INCREQ(giCmdReqRemoved))
        RemoveResponseFile( giCmdReqRemoved );

    DBG((FL,"\t2: giCmdReq=%d, giCmdResp=%d, giCmdReqRemoved=%d\n", 
            giCmdReq, giCmdResp, giCmdReqRemoved));
        
    // Now: giCmdResp == giCmdReqRemoved

    //
    // Build a new one 
    //FILENAME_FORMAT
    if ( iParCommOut == PARAM_COMM_OUT_FILE )
        sprintf(szRespFilename, FILENAME_FORMAT, //"pymidas%2.2s%1d.%04d", 
            gszMidasUnit, giCid, giCmdReq);
    else // PARAM_COMM_OUT_FILETERM
        sprintf(szRespFilename, FILENAME_FORMAT"+terminal", //"pymidas%2.2s%1d.%04d+terminal",
            gszMidasUnit, giCid, giCmdReq);
    INCREQ(giCmdReq);
    
    sprintf(szFilename,"%s%s",getenv("MID_WORK"),szRespFilename);

    DBG((FL,"\t%s\n", szFilename));
    DBG((FL,"<-- GetNewResponseFileNameReq() - end\n"));
    
	//return gszRespFilename;
	return szFilename;
}//GetNewResponseFileNameReq()
//------------------------------------------------------------------------
// GetNewResponseFileNameResp()
//
// Build and returns response file name for response
//
char * GetNewResponseFileNameResp()
{
    static char szRespFilename[100];
    static char szFilename[100]; 
    int iParCommOut  = ReadParamInt(PAR_COMM_OUT); 

    DBG((FL,"==> GetNewResponseFileNameResp() - begin\n"));
    //if ( iParCommOut == PARAM_COMM_OUT_FILE )
    //sprintf(szRespFilename, "pymidas%03d.%04d", giCid, giCmdResp++);
    DBG((FL,"\tgiCmdReq=%d, giCmdResp=%d, giCmdReqRemoved=%d\n", 
            giCmdReq, giCmdResp, giCmdReqRemoved));
    sprintf(szRespFilename, FILENAME_FORMAT, 
        gszMidasUnit, giCid, giCmdResp);
    INCREQ(giCmdResp);    
    
    //else // PARAM_COMM_OUT_FILETERM
    //    sprintf(szRespFilename, "pymidas%03d.%04d+terminal", giCid, giCmdResp++);
    
    sprintf(szFilename,"%s%s",getenv("MID_WORK"),szRespFilename);
    DBG((FL,"\t%s\n", szFilename));
    DBG((FL,"<-- GetNewResponseFileNameResp() - end\n"));
    return szFilename;
}//GetNewResponseFileNameResp()

//------------------------------------------------------------------------
// IsAsyncCmd()
//
// If the cmd is MIDAS async command this will return true otherwise false
//
bool IsAsyncCmd(char *pszCmd)
{
	bool boRet = false;
	
	if ( strncmp(pszCmd,"CHANGE/DIRECTORY",10) == 0 )
		return true;
	if ( strncmp(pszCmd,"CLEAR/BACKGROUND",10) == 0 )
		return true;
	
	return boRet;
}//IsAsyncCmd()

//------------------------------------------------------------------------
int ReadParamInt(char *pszParamName)
{
	int iRet = 0;
	
    //printf("1: ReadParamInt()=%d\n", iRet);
	for ( int i = 0; aParams[i].szName[0] != '\0'; i++ )
	{
		if ( strcmp(aParams[i].szName, pszParamName) == 0 )
		{	    		
			iRet = atoi(aParams[i].szValue);
//		    try 
//		    {
//		    	iRet = atoi(aParams[i].szValue);
//			} 
//		    catch (Exception e) 
//		    {
//		      	iRet = 0;
//			}			
		}//if
	}//for 

    //printf("2: ReadParamInt()=%d\n", iRet);
	return iRet;
}//ReadParamInt()

//------------------------------------------------------------------------
// main()
// 
// This function is just for testing purposes.
//
int main(int argc, char **argv) 
{ 
	int  stat, jj; 
    char wstr[160];
    int  iLen=3; 
    char cType;
      
    stat = OSY_TRNLOG("MID_WORK",wstr,160,&jj);    /* Decode startup directory */
    printf("stat=%d  jj=%d  wstr=%s\n", stat, jj, wstr);  

    strcpy(wstr,"                                               ");
    stat = PymidasGetKeywordValue("MID$ERRMESS", wstr, &cType, &iLen);
    printf("stat=%d  wstr='%s' type='%c' len=%d\n", 
        stat , wstr, cType, iLen);  
    
    strcpy(wstr,"                                               ");
    stat = PymidasGetKeywordValue("MID$SYS", wstr, &cType, &iLen);
    printf("stat=%d  wstr='%s' type='%c' len=%d\n", 
        stat , wstr, cType, iLen);  
    
    strcpy(wstr,"                                               ");
    stat = PymidasGetKeywordValue("USER", wstr, &cType, &iLen);
    printf("stat=%d  wstr='%s' type='%c' len=%d\n", 
        stat , wstr, cType, iLen);  

    int iKeyw = 0;
    stat = PymidasGetKeywordValue("LOG", wstr, &cType, &iLen);
    printf("stat=%d  wstr='%s' type='%c' len=%d\n", 
        stat , wstr, cType, iLen);  

//finito:
    // MIDAS Epilog will stop the application, so this should be the last 
    // call in the main routine of your MIDAS application
    //
    SCSEPI(); 
}//main

//------------------------------------------------------------------------
// PymidasGetKeywordValue()
//
// Get the value of the given keyword 
//
// in       pszKeyName  Name of the keyword
// out      pValue      Value of the requested keyword 
// out      pcType      Type of the keyword
// out      piActLen    Actual length of the value: 
//                      pcType=='C' - length of the string
//                      other - # of elements in an array
// returns  int         Error code:
//          0           No error
int  PymidasGetKeywordValue(
    char *pszKeyName, 
    void *pValue, 
    char *pcType, 
    int  *piActLen)
{
    //char *pszValue = (char*) pValue;
    char  cType    = '\0';                 // type of key
    int   iNoElems = 0;                    // no. of elements
    int   iBytes   = 0;                    // no. of bytes per element
    int   iRet     = 0;
    int   iReturn  = PYMIDAS_ERR_NOERROR;  // return value of this function
        
    DBG((FL,"==> PymidasGetKeywordValue() - begin\n"));

    LayerProlog();
 
 	iRet = Initialize("00", "00", 0);
    DBG((FL, "\tA1: Ret=%d\n", iRet ));
 
    *piActLen = 0;
   
    DBG((FL, "\tA2: KeyName=%s,\n",pszKeyName));
    DBG((FL, "\tA3: Type='%c'\n", *pcType ));
    DBG((FL, "\tA4: NoElem=%d\n",iNoElems));
    DBG((FL, "\tA5: Bytes=%d\n",iBytes ));
    
    iRet = SCKFND(pszKeyName, pcType, &iNoElems, &iBytes);
    DBG((FL, "\tB: Ret=%d, KeyName=%s, Type='%c', NoElem=%d, Bytes=%d\n",
        iRet, pszKeyName, *pcType, iNoElems, iBytes ));
    
    if ( *pcType == PYMIDAS_KEYWORD_BLANK )
    {
        iReturn = PYMIDAS_ERR_KEYW_NOT_FOUND;
        DBG((FL,"\t ERROR %d: PYMIDAS_ERR_KEYW_NOT_FOUND\n", iReturn ));
        goto finito;
    }//if
    //
    // Found something
    //
    if ( *pcType == PYMIDAS_KEYWORD_CHAR )
    {
        int  iActvals = 0;
        char *pszValue = (char*) pValue;
        int  iMax = strlen(pszValue);
        char szBuf[200];
        
        DBG((FL,"\t PYMIDAS_KEYWORD_CHAR iMax=%d\n", iMax));
        
        if ( SCKGETC(pszKeyName, 1, 199, &iActvals, szBuf) )
        {
            iReturn = PYMIDAS_ERR_KEYW_VALUE;
            DBG((FL,"\t ERROR %d: PYMIDAS_ERR_KEYW_VALUE\n", iReturn ));
            goto finito;
        }//if 
        DBG((FL,"\t pszKeyName=%s: '%s' iActvals=%d max=%d\n", 
            pszKeyName, szBuf, iActvals, iMax ));
        strcpy(pszValue, szBuf);
        *piActLen = iActvals;
    }//if
    else if ( *pcType == PYMIDAS_KEYWORD_INTEGER )
    { 
        int iActvals    = 0;
        int iElem       = 0;
        int iValue      = 0;
        int  iUnit      = 0;
        int  iNull      = 0;
        char *pszBase = (char*) pValue;
        char *pszValue = pszBase;
        
        *pszValue = '\0';
        
        DBG((FL,"\t PYMIDAS_KEYWORD_INTEGER elems=%d\n", iNoElems ));
        for ( iElem = 1; 
                iElem <= iNoElems; 
                    iElem++, pszValue = pszBase+strlen(pszBase)) 
        {
            if ( SCKRDI(pszKeyName, iElem, 1, &iActvals, &iValue, &iUnit, &iNull) )
            {
                iReturn = PYMIDAS_ERR_KEYW_VALUE;
                DBG((FL,"\t ERROR %d: PYMIDAS_ERR_KEYW_VALUE\n", iReturn ));
                goto finito;
            }//if 
            if ( iElem > 1 ) 
                *pszValue++ = ',';        // separator
                 
            sprintf(pszValue, "%d", iValue);
        }//for
        *piActLen = iNoElems;
        DBG((FL,"\t[%2d]=%3d str=%s<\n", iElem, iValue, pszBase));
    }//else if
    else
    {
        iReturn = PYMIDAS_ERR_NOT_SUPPORTED;
        DBG((FL,"\t ERROR %d: PYMIDAS_ERR_NOT_SUPPORTED", iReturn ));
        goto finito;
    }//else
    
finito: 
    DBG((FL,"<-- PymidasGetKeywordValue() - end (%d)\n", iReturn));
    return iReturn;
}//PymidasGetKeywordValue

//-----------------------------------------------
//--- PymidasGetKeywordValue() Python wrapper ---
//
// value,type,len=PyMidas.GetKeywordValue('<keyword>')
//
static PyObject * wrap_PymidasGetKeywordValue(PyObject *self, PyObject *args)
{
    char *pszKeyName; //const char *pszKeyName;
    char *pcType = NULL;
          char  cType;
    const char *pValue;
    int iActLen=0;
    int iLen = 0; 
    int iResult;
    int iRet;

    DBG((FL,"================================================================\n"));
    DBG((FL,"wrap_PymidasGetKeywordValue() -begin \n")); 

    if (!PyArg_ParseTuple(args, "s", &pszKeyName))
    {
        DBG((FL,"\tPyArg_ParseTuple() *ERROR*\n"));
        return NULL;    
    }
    DBG((FL,"\tPyArg_ParseTuple() OK\n"));
    DBG((FL,"\t1: pszKeyName=%s.\n", pszKeyName ));

    char szValue[200] = "";
    iResult = PymidasGetKeywordValue( pszKeyName, 
        szValue, 
        &cType, 
        &iLen);
    DBG((FL,"\tPymidasGetKeywordValue() return=%d.\n", iResult));
    
    if ( iResult )
    {
        DBG((FL,"\t ... FAILED!\n"));
        goto finito;
    }//if
    
    DBG((FL,"\t2: pszKeyName=%s, value=%s, pcType=%c, iLen=%d.\n",
            pszKeyName, szValue, cType, iLen ));

finito:
    DBG((FL,"wrap_PymidasGetKeywordValue() - end (%d)\n", iResult));
    return Py_BuildValue("sci",szValue, cType, iLen);
}//wrap_function()

//------------------------------------------------------------------------
// Python wrapper
//------------------------------------------------------------------------,
#ifdef __cplusplus
extern "C" {
#endif
int function (int x); // Test function for Python wrapper
#ifdef __cplusplus
}
#endif 

  
//------------------------------------------------------------------------
// Testing, testing, ...
//
int function (int x) // Test function for Python wrapper
{
    DBG((FL,"Python wrapper -> function(): x=%d\n", x));
    return x+1;
}//function()
//---
static PyObject * wrap_function(PyObject *self, PyObject *args)
{
    const char *command;
    int result;
    int x;
    
    if (!PyArg_ParseTuple(args, "i" ,&x))
    //if (!PyArg_GetInt(args,&x))
        return NULL;
    result = function(x);
    return Py_BuildValue("i",result);
}//wrap_function()
 
//------------------------------------------------------------------------
// PymidasConnection()
//
// returns  int         Error code:
//          0           No error
int PymidasConnection( char *pszCmd )
{
    int iReturn = 0;
    static char szMidasunit[] = "33";
    static char szHost[] = "orion.astro.helsinki.fi";
    static int siCid = 0;
    
    DBG((FL,"PymidasConnection( () - begin\n"));

    LayerProlog();

    if ( strcmp(pszCmd,"OPEN") == 0 || strcmp(pszCmd,"open") == 0 )
    {
        DBG((FL,"\tOPEN found!\n"));
        iReturn = XCXOPN(szMidasunit, szHost, &siCid);
        DBG((FL,"\tXCXOPN() done, siCid=%d urn=%d\n", siCid, iReturn));
    }//if
    else 
    {
        DBG((FL,"PymidasConnection( () - begin\n"));
    }//else
    
finito: 
    DBG((FL,"PymidasConnection() - end (%d)\n", iReturn));
    return iReturn;
}//PymidasConnection
//
//--- Pymidas_Connection() Python wrapper ---
//
// value,type,len=PyMidas.GetKeywordValue('<keyword>')
//
static PyObject *wrap_PymidasConnection(PyObject *self, PyObject *args)
{
    char *command;
    int sts;

    DBG((FL,"================================================================\n"));
    DBG((FL,"wrap_PymidasConnection() -begin\n"));

    if (!PyArg_ParseTuple(args, "s", &command))
        return NULL;
    sts = PymidasConnection(command); 
    DBG((FL,"wrap_PymidasConnection() - end (%d)\n", sts));
    return Py_BuildValue("i", sts);
}

//------------------------------------------------------------------------
#define DEC    ( void *(*)(void*) )
void *threadwork(void *p)
{
    /*
    int iReturn = 0;
    int iRet = 0;
    char *args[] = { 
        "/home/toittine/midas/04SEPpl1.0/monit/qreqa.exe", 
        "READ/KEYW user",
        NULL};

    DBG((FL,"==> threadwork() - begin\n"));
    try 
    {
        iRet = execl("/home/toittine/midas/04SEPpl1.0/monit/qreqa.exe", args, NULL);
        DBG((FL,"\tiRet=%d, Error: (%d) %s\n", iRet, errno, strerror(errno)));
	} 
    catch (Exception e) 
    {
        DBG((FL,"\tiRet=%d, Error: (%d) %s\n", iRet, errno, strerror(errno), e));
	}
    DBG((FL,"<-- threadwork() - end (0x%08X)\n", iReturn));
    */
}

//------------------------------------------------------------------------
void SignalTrapper(int iSig)
{
    DBG((FL,"==> SignalTrapper(%d) - begin\n", iSig));  
#if OSTYPE_LINUX
    // Nothing to do so far.
    sighandler_t xSig = signal(2, SignalTrapper);        
    DBG((FL,"\t xSig == 0x%X (SIG_ERR=0x%X) SIGKILL=0x%X SIGSTOP=0x%X SIGUSR2=0x%X\n", 
        xSig, SIG_ERR, SIGKILL, SIGSTOP, SIGUSR2));      
    if ( xSig == SIG_ERR ) 
    {
        DBG((FL,"\t xSig == SIG_ERR (0x%X)\n", xSig));      
    }//if
#endif
    DBG((FL,"<-- SignalTrapper() - end\n"));  
}//SignalTrapper

#define MAX_TOKENS          20
#define MAX_TOKEN_LEN       80

//------------------------------------------------------------------------
// PymidasDo()
//
// Works like a 'drs'. Executes generic MIDAS commands
//
// in       pszCmd      MIDAS command to execute ("READ/KEYW","LOAD/IMA",...)
// out      pResp       Response string from MIDAS.
// returns  int         Error code:
//          0           No error

int PymidasDo( char *pszCmd, char **pResp )
{
    int iReturn = 0;
    int iRet = 0;
    static int cid = 0;
//    static int iCmd = 0;
    int retstat[2];
    char szMainCmd[1000];
    //char szRespFilename[100]; 
    //static char szRespString[10000];
    static char *pszRespString;
    char szEnv[80];
    long lReadCount = 0L;
    int iParMidasCmdTimeout = ReadParamInt(PAR_MIDAS_CMD_TIMEOUT);
    //int iParCommMode = ReadParamInt(PAR_COMM_MODE);
    
    DBG((FL,"==> PymidasDo() - begin\n"));  
    DBG((FL,"\tCmd='%s' InitDone=%d MID_WORK='%s'\n", 
        pszCmd, giInitDone, getenv("MID_WORK"))); 
   
    // Return an empty string if failure
    //memset(szRespString,'\0',10000);
     
    *pResp = pszRespString;


    // Response file is format "pymidas001.0002" where
    // 001  - MIDAS session ID
    // 0002 - request number for this response
    //sprintf(szRespFilename, "pymidas%03d.%04d", giCid, iCmd++);
    //sprintf(szMainCmd, "%s >%s", pszCmd, szRespFilename);
    
    //sprintf(gszRespFilename, "pymidas%03d.%04d", giCid, iCmd++);
    sprintf(szMainCmd, "%s >%s", pszCmd, GetNewResponseFileNameReq() );
    //sprintf(szMainCmd, "%s >%s", pszCmd, GetResponseFileName(giCid, iCmd++);
	
	//
	// Send command to MIDAS but if the command is an asynchronous don't
	// wait. Otherwise use the value of parameter.
	//
	int iWaitTime = (IsAsyncCmd(szMainCmd) ? 0 : iParMidasCmdTimeout );

	DBG((FL,"\t XCCSND(%d,'%s',%d,...)\n",giCid, szMainCmd,iWaitTime));
    DBG((FL,"\t\t cmd len=%d\n",strlen(szMainCmd)));
 
	// Wait as long as parameter iWaitTime said: 0=no wait, -1=forever, ...
    iRet = XCCSND(giCid, szMainCmd, iWaitTime, retstat);
    
//    if ( iParCommMode == PARAM_COMM_MODE_FILE )
//        iRet = XCCSND(giCid, szMainCmd, iWaitTime, retstat);
//    else if ( iParCommMode == PARAM_COMM_MODE_SOCKET )
//        iRet = XCCSND(giCid, szMainCmd, iWaitTime, retstat);
//    else
//    { 
//        iReturn = PYMIDAS_ERR_NOT_SUPPORTED;
//        goto finito;
//    }//else       
    DBG((FL,"\t XCCSND(%d,'%s',%d,..)=%d \n\trestat=0x%08X 0x%08X\n", 
           giCid, szMainCmd, iWaitTime, iRet,retstat[0],retstat[1]));
    
    if ( retstat[0] ) 
    {
        //Error!
        DBG((FL,"\t MIDAS ERROR CODE 0x%08X\n", retstat[0] ));
        iReturn = retstat[0];
    }//if
    else
    {
    	if ( iWaitTime == 0 )  // This is an async cmd
        {
            DBG((FL,"\tgoto finito\n"));
    		goto finito;
        }	  
        // Ok, so read the response
        DBG((FL,"\t Read response from file '%s'.\n", gszRespFilename));
        
        lReadCount = ReadResponse(&pszRespString, &iReturn);         
   }//else
    
    if ( strcmp( pszCmd, "BYE" ) == 0 )
    {
        // It's time to say goodbye
        DBG((FL,"\tBYE! -->\n"));        
        iRet = XCXCLO(giCid);
        DBG((FL,"\t XCXCLO()=%d cid=%d\n", iRet,giCid));        
        giInitDone = 0;
    }//if
        
finito:   
    DBG((FL,"<-- PymidasDo() - end (0x%08X,0x%08X)\n", iReturn,pszRespString));
    return iReturn;
}//PymidasDo

 
//------------------------------------------------------------------------
// ReadResponseFile()
//
// Reads the response from given file and gives a static pointer to it.
//
// in       fp      	File pointer
// out      ppResp      Pointer to pointer to response string.
// returns  long        Size of the response file
//          0           No response or the file was empty.
//
long ReadResponseFile( FILE *fp, char **ppResp )
{
	//static char *pBuf;
	//static long lBufSize;
	long lFileSize = 0L;
	long lReads = 0L;
	
    DBG((FL,"==> ReadResponseFile() - begin\n"));  
    DBG((FL,"\tfp=%p glBufSize=%ld\n",  fp, glBufSize));

	if ( fp != NULL )
	{
		// Find size of the file:
		fseek( fp, 0L, SEEK_END);   //to the end of the file 
		lFileSize = ftell( fp );	//where is it?
	    DBG((FL,"\t\tlSize=%d\n",  lFileSize));
		fseek( fp, 0L, SEEK_SET);	//to the beginning of the file
		//
		// Allocate new buffer if needed
		//
		if ( lFileSize >= glBufSize )
		{
			// Need a new buffer
			glBufSize = lFileSize + 1000L;  // new size is +1kB
		    DBG((FL,"\t\t new Buffer glBufSize=%d\n",  glBufSize));
			if ( gpResponseBuf != NULL ) 
				delete[] gpResponseBuf;
			gpResponseBuf = new char[ glBufSize ];
		}//if
	}//if
	
	if ( gpResponseBuf == NULL )
	{
		gpResponseBuf = ""; // will return something
		lFileSize = glBufSize = 1L;
		goto finito;
	}//if
	memset( gpResponseBuf ,'\0', glBufSize );
	//
	// Read actual data
	//
    DBG((FL,"\t pBuf=%p , lBufSize=%ld , ftell(fp)=%ld \n", 
    	gpResponseBuf, glBufSize, ftell(fp)));
    
    //lReads = fread( gpResponseBuf, lFileSize, 1, fp );
    //DBG((FL,"\tlReads=%ld\n", lReads));
    lReads = 0L;
    while ( (lReads = fread( gpResponseBuf+lReads, lFileSize, 1, fp )) > 0L )
    {
        DBG((FL,"\tlReads=%ld\n", lReads));
    	//sleep(1);
    }//while

	//
	// Build return parameter
	// 
finito:
	*ppResp = gpResponseBuf;
    //DBG((FL,"<-- ReadResponseFile() - end (%ld)\n", lReads));
    //return (lReads != 0L ? lFileSize : 0L );
    DBG((FL,"<-- ReadResponseFile() - end (%ld)\n", lFileSize));
	return (lFileSize);
}//ReadResponseFile

//------------------------------------------------------------------------
// KeywMidmodeRead()
//
int KeywMidmode10Read()
{
    int  iRet = 0;
    int  iActvals = 0;
    int  iValue = 0;
    int  iUnit = 0;
    int  iNull = 0;
    int  iKstat = 0;

    DBG((FL,"==> KeywMidmode10Read() - begin\n"));  
    iValue = 0; // Initialize
    //
    // MIDAS will notice a new command defining/deleting via MID$MODE(10)
    //
    iRet = XCKRDI( 
             giCid,      //IN: connection id of background Midas
             "MID$MODE", //IN: keyword name (max. 8 chars.)
             10,         //IN: first data item to be read
             1,          //IN: no. of elements to get
             &iActvals,  //OUT: actual no. of elements returned
             &iValue,   //OUT: buffer for data values
             &iUnit,     //OUT: address of unit-pointer
             &iNull,     //OUT: no. of null values in keyword
             &iKstat     //OUT: status from SCKRDI in background Midas
             );
    if ( iRet == 0 )
    {
        if ( iActvals == 1 )
        {
            if ( iValue == 1 )
                iRet = PYMIDAS_RET_CMDDEFINED;
            else if ( iValue == 2 )
                iRet = PYMIDAS_RET_CMDDELETED;
            else
                iRet = PYMIDAS_RET_OK;
        }
        else
        {
            iRet = PYMIDAS_ERR_GENERAL;
        }
    }
    else 
    {
        iRet = PYMIDAS_ERR_GENERAL;
    }
    DBG((FL,"<-- KeywMidmode10Read() - end (%d) \n", iRet));
    return (iRet);
}//KeywMidmode10Read
//------------------------------------------------------------------------
// KeywMidmodeReset()
//
int KeywMidmode10Reset()
{
    int  iRet = 0;
    int  iActvals = 0;
    int  iValue = 0;
    int  iUnit = 0;
    int  iNull = 0;
    int  iKstat = 0;

    DBG((FL,"==> KeywMidmode10Reset() - begin\n"));  

    iRet = XCKWRI( 
             giCid,      //IN: connection id of background Midas
             "MID$MODE", //IN: keyword name (max. 8 chars.)
             &iValue,   //IN: buffer with data values
             10,         //IN: first data item to be written
             1,          //IN: no. of elements to write
             &iUnit,     //IN: unit-pointer
             &iKstat     //OUT: status from SCKRDI in background Midas
             );
    if ( iRet == 0 )
    {
        iRet = PYMIDAS_RET_OK;
    }
    else 
    {
        iRet = PYMIDAS_ERR_GENERAL;
    }
    DBG((FL,"<-- KeywMidmode10Reset() - end (%d) \n", iRet));
    return (iRet);
}//KeywMidmode10Reset

//------------------------------------------------------------------------
// ReadResponse()
//
// Reads the response and gives a static pointer to it.
// We also need to check if MIDAS has changed cmd database for some reason.
//
// out      ppResp      Pointer to pointer to response string.
//          piRet       New return value of the layer:
//                          0 - OK
//                         99 - MIDAS_RETSTAT_CMDDEFINED 
//                         98 - MIDAS_RETSTAT_CMDDELETED
// returns  long        Size of the response file
//          0           The file was empty.
//			PYMIDAS_ERR_NORESPONSE          0xF0010010
//						No response file
//
long ReadResponse( char **ppResp, int *piRet )
{
	long lReturn = 0L;
	char szFileName[100];
    int  iRet = 0;
    int  iMidMode = 0;
    
    DBG((FL,"==> ReadResponse() - begin\n"));  
    strcpy(szFileName, GetNewResponseFileNameResp());
    DBG((FL,"\t RespFilename=%s.\n", szFileName));

	FILE *fp =  fopen(szFileName,"rt");
	if ( fp == NULL )
	{
	    DBG((FL,"\t\t No response file! ***\n"));
	    lReturn = 0L; //PYMIDAS_ERR_NORESPONSE;
	    *ppResp = gpResponseBuf;
	}//if
	else
	{
	    DBG((FL,"\t\t File is open ... \n"));
	    //int iReadCount = fread( szRespString, 1000, 1, fp );
	    lReturn = ReadResponseFile( fp, ppResp); //&pszRespString );
	    DBG((FL,"\t\t ... and size=%ld \n", lReturn));            
		//*pResp = pszRespString;
	    //int iReadCount = 
	    //fread( pszRespString, lReadCount, 1, fp );
	    //
	    // Delete the response file
	    //
	    fclose(fp);
	    remove(gszRespFilename);            
	}//else
    //
    // MIDAS will notice a new command defining/deleting via MID$MODE(10)
    //
    iRet = KeywMidmode10Read();  // &iMidMode );
//    iRet = XCKRDI( 
//             giCid,      //IN: connection id of background Midas
//             "MID$MODE", //IN: keyword name (max. 8 chars.)
//             10,         //IN: first data item to be read
//             1,          //IN: no. of elements to get
//             &iActvals,  //OUT: actual no. of elements returned
//             &iValues,   //OUT: buffer for data values
//             &iUnit,     //OUT: address of unit-pointer
//             &iNull,     //OUT: no. of null values in keyword
//             &iKstat     //OUT: status from SCKRDI in background Midas
//             );
    *piRet = iRet;
    if ( iRet == PYMIDAS_RET_CMDDEFINED || iRet == PYMIDAS_RET_CMDDELETED )
    {
        iRet = KeywMidmode10Reset();
    }//if

//    if ( iRet == 0 )
//    {
//        if ( iActvals == 1 )
//        {
//            if ( iValues == 1 )
//            {
//                // Reset the keyword
//                //
//                
//                
//                *piRet = PYMIDAS_RET_CMDDEFINED;
//            }
//            else if ( iValues == 2 )
//                *piRet = PYMIDAS_RET_CMDDELETED;
//            else
//                *piRet = PYMIDAS_RET_OK;
//        }
//        else
//            *piRet = PYMIDAS_ERR_GENERAL;
//    }
//    else 
//    {
//        *piRet = PYMIDAS_ERR_GENERAL;
//    }
    //
    // Final
    //
    DBG((FL,"<-- ReadResponse() - end (%ld) ret=%d\n", lReturn,*piRet));
    return lReturn;
}//ReadResponse()

//-----------------------------------
//--- Pymidas_Do() Python wrapper ---
//
// value,type,len=PyMidas.GetKeywordValue('<keyword>')
//
static PyObject *wrap_PymidasDo(PyObject *self, PyObject *args)
{
    char *pszCmd;
    int iReturn = 0;
    char *pResp = NULL;

    DBG((FL,"================================================================\n"));
    DBG((FL,"==> wrap_PymidasDo() - begin\n"));

    if ( !PyArg_ParseTuple(args, "s", &pszCmd) )
    {
        DBG((FL,"\tPyArg_ParseTuple() *ERROR*\n"));
        iReturn = PYMIDAS_ERR_WRAP_ARG;
    }//if  

    DBG((FL,"\tCmd='%s'\n", pszCmd)); 
    iReturn = PymidasDo(pszCmd, &pResp);
    
//    if ( (iReturn = PymidasDo(pszCmd)) != PYMIDAS_ERR_NOERROR )
//    {
//        DBG((FL,"\PymidasDo() return 0x%08X\n", iReturn));
//        iReturn = PYMIDAS_ERR_GENERAL;
//    }//if 
    
   
finito:
    DBG((FL,"<-- wrap_PymidasDo() - end (0x%08X,0x%08X)\n",iReturn,pResp));
    return Py_BuildValue("is", iReturn, pResp);
}//wrap_PymidasDo

//-----------------------------------
// Initialize()
//
//
int Initialize(char *pszMidasUnit, char *pszMyUnit, int iCid)
{
	int  stat, jj;
	char   wstr[160];
	char  *cp; 
	char  cxp[4];
	int   ip[6],fid;
	register int  nr;
    int iReturn = 0;
    int iParCommMode = ReadParamInt(PAR_COMM_MODE);
    int iParCommOut  = ReadParamInt(PAR_COMM_OUT);

    DBG((FL,"==> Initialize() - begin\n"));
    DBG((FL,"\tCOMM_MODE=%d (%s), COMM_OUT=%d (%s)\n", 
        iParCommMode, iParCommMode==0?"files":"sockets", 
        iParCommOut,  iParCommOut==0 ?"file" :"file+terminal"));
    
    //gpszMidasUnit = pszMidasUnit; 
    //gpszMyUnit = pszMyUnit;
    giCid = iCid;
    
    for (jj=0; jj<4; jj++) cxp[jj] = 'N';
    
    /*  get symbols DAZUNIT, MID_WORK (as was done in INPREPA before)  */
    
    OSY_GETSYMB("DAZUNIT",wstr,4);
    gszMidasUnit[0] = FRONT.DAZUNIT[0] = wstr[0];
    gszMidasUnit[1] = FRONT.DAZUNIT[1] = wstr[1];
	gszMidasUnit[2] = '\0';

	gszMyUnit[0] = '0';
	gszMyUnit[1] = '0';
	gszMyUnit[2] = '\0';

    DBG((FL,"\tFRONT.DAZUNIT[0]=%c.\n",FRONT.DAZUNIT[0]));
    DBG((FL,"\tFRONT.DAZUNIT[1]=%c.\n",FRONT.DAZUNIT[1]));
    
    stat = OSY_TRNLOG("MID_WORK",wstr,160,&jj);    /* Decode startup directory */
    DBG((FL,"\tstat=%d jj=%d\n",stat,jj));
    if ( (stat != 0) || (jj > 160) )
    {
	    DBG((FL,"\tERROR!\n"));
       (void) printf
       ("we could not translate MID_WORK or MID_WORK > 160 char.\n\r");
       //ospexit(1);
    }
   	if (wstr[jj-1] != FSY_DIREND)   
    {
       wstr[jj++] = FSY_DIREND;
       wstr[jj] = '\0';
    }//if

    (void) strcpy(FRONT.STARTUP,wstr);  /* save name of startup directory */
    //(void) strcpy(FRONT.STARTUP,wstr);  /* save name of startup directory */
    DBG((FL,"\t start up directory=%s.\n",wstr));
    
    (void) strcpy(FRONT.PROMPT,"Midas ");   /* init prompt string  */
    FRONT.PEND = 6;                         /* length of prompt + 1 */

    if ( !giInitDone )
    {
        // Open MIDAS in a background mode
        //iRet = XCZINI("66", "60", "", "", "", 10, &cid);
        //DBG((FL,"\t XCZINI()=%d cid=%d\n", iRet,cid));

		if ( iParCommMode == PARAM_COMM_MODE_FILE )
		{
	        DBG((FL,"\t F i l e s ! \n")); 
	        DBG((FL,"\t XCZOPN('%s','%s','\\0',%d)\n", 
	            gpszMidasUnit, gpszMyUnit,giCid));      

	        iReturn = XCZOPN(gpszMidasUnit, gpszMyUnit, "", &giCid);

	        DBG((FL,"\t XCZOPN('%s','%s','\\0',%d)=%d\n", 
	            gpszMidasUnit, gpszMyUnit,giCid,iReturn));      
		}//if
		else if ( iParCommMode == PARAM_COMM_MODE_SOCKET )
		{
	        DBG((FL,"\t S o c k e t s ! \n")); 
	        DBG((FL,"\t XCXOPN('%s','%s','\\0',%d)\n", 
	            gpszMidasUnit, gpszMyUnit,giCid));      

	        iReturn = XCXOPN(gpszMidasUnit, "", &giCid);

	        DBG((FL,"\t XCXOPN('%s','%s','\\0',%d)=%d\n", 
	            gpszMidasUnit, gpszMyUnit,giCid,iReturn));      
		}//else if
		else
		{
			iReturn = PYMIDAS_ERR_GENERAL;
		}//else
		
        // MIDAS sends some nice signals. Try to trap those.
        //
#if OSTYPE_LINUX
        sighandler_t xSig = signal(SIGUSR2, SignalTrapper);        
        DBG((FL,"\t xSig == 0x%X (SIG_ERR=0x%X) SIGKILL=0x%X SIGSTOP=0x%X SIGUSR2=0x%X\n", 
            xSig, SIG_ERR, SIGKILL, SIGSTOP, SIGUSR2));      
        if ( xSig == SIG_ERR ) 
        {
            DBG((FL,"\t xSig == SIG_ERR (0x%X)\n", xSig));      
        }//if
        // MIDAS sends some nice signals. Try to trap those.
        /*
        xSig = signal(2, SignalTrapper);        
        DBG((FL,"\t xSig == 0x%X (SIG_ERR=0x%X) SIGKILL=0x%X SIGSTOP=0x%X SIGUSR2=0x%X\n", 
            xSig, SIG_ERR, SIGKILL, SIGSTOP, SIGUSR2));      
        if ( xSig == SIG_ERR ) 
        {
            DBG((FL,"\t xSig == SIG_ERR (0x%X)\n", xSig));      
        }//if
        */
#endif      
        giInitDone++;
    }//if

	//qinit_here();
	
    //  get PID + also store PID in file MIDASxy.PID      
    //FRONT.PID = oshpid();           //  get process id 

	//giInitDone = 1;
	giCid = 0;
	
    DBG((FL,"\t MidasUnit='%s', MyUnit='%s', Cid=%d\n", 
        gpszMidasUnit,gpszMyUnit,giCid)); 

    DBG((FL,"<-- Initialize() - end (0x%08X)\n", iReturn));
    return iReturn; 
}//Initialize

//-----------------------------------
//--- Pymidas_Do() Python wrapper ---
//
// value,type,len=PyMidas.GetKeywordValue('<keyword>')
//
static PyObject *wrap_Initialize(PyObject *self, PyObject *args)
{
    char *pszMidasUnit = NULL;
    char *pszMyUnit    = NULL;
    int  iCid = 0;
    int iReturn = 0;

    DBG((FL,"================================================================\n"));
    DBG((FL,"==> wrap_Initialize() - begin\n"));

    if ( !PyArg_ParseTuple(args, "ssi", &pszMidasUnit, &pszMyUnit, &iCid) )
    { 
        DBG((FL,"\tPyArg_ParseTuple() *ERROR*\n"));
        iReturn = PYMIDAS_ERR_WRAP_ARG;
    }//if

    DBG((FL,"\t MidasUnit='%s', MyUnit='%s', Cid=%d\n", 
        pszMidasUnit,pszMyUnit,iCid)); 

    iReturn = Initialize(pszMidasUnit, pszMyUnit, iCid);

finito:
    DBG((FL,"<-- wrap_Initialize() - end (0x%08X)\n", iReturn));
    return Py_BuildValue("i", iReturn);
}//wrap_Initialize

//------------------------------------------------------------------------
// SetParam()
//
// Set a parameter value.
//
// in       pszParamName	Name of the parameter
// in		pszParamValue	New value
// returns  int        		Error code
//          0           	No error.
//			PYMIDAS_ERR_NOT_SUPPORTED 		0xF0010001
//				If parameter name was not recognized
//			PYMIDAS_ERR_GENERAL             0xF0010002
//				If initialization of connection has been done already
//
int SetParam(char *pszParamName, char *pszParamValue)
{
    int iReturn = 0;
	bool boFound = false;
	
    DBG((FL,"==> SetParam() - begin\n"));
    
    DBG((FL,"\t ParamName='%s', ParamValue='%s'\n",pszParamName, pszParamValue)); 

	if ( strcmp(pszParamName, PAR_COMM_MODE) == 0 )
	{
		// If we has made initialization we can't change connection mode
		//
	    if ( giInitDone )
	    {
    		DBG((FL,"\t *** ERROR: Init already done! (PYMIDAS_ERR_GENERAL)\n"));
	    	iReturn = PYMIDAS_ERR_GENERAL;
	    	goto finito;
	    }	
	}//if

	for ( int i = 0; aParams[i].szName[0] != '\0'; i++ )
	{
		if ( strcmp(aParams[i].szName, pszParamName) == 0 )
		{			
			// This is the right one
	    	DBG((FL,"\t %s: '%s'->'%s'   %s\n",
	    		pszParamName, aParams[i].szValue, pszParamValue,
	    		(aParams[i].iType==0 ? "STRING":"INTEGER"))); 
	    		
	    	strcpy(aParams[i].szValue, pszParamValue);
			boFound = true;
		}//if
	}//for 

	if ( ! boFound )
	{
    	DBG((FL,"\t *** ERROR: Parameter not found! (PYMIDAS_ERR_NOT_SUPPORTED)\n"));
    	iReturn = PYMIDAS_ERR_NOT_SUPPORTED;
	}//if

finito:
    DBG((FL,"<-- SetParam() - end (0x%08X)\n", iReturn));
    return iReturn;
}//SetParam

//--------------------------------------
//--- wrap_SetParam() Python wrapper ---
//
// ret = PyMidas.SetParam '<param_name>' '<value>'
//
static PyObject *wrap_SetParam(PyObject *self, PyObject *args)
{
    char *pszParamName  = NULL;
    char *pszParamValue = NULL;
    int iReturn = 0;

    DBG((FL,"================================================================\n"));
    DBG((FL,"==> wrap_SetParam() - begin\n"));

    if ( !PyArg_ParseTuple(args, "ss", &pszParamName, &pszParamValue) )
    {
        DBG((FL,"\tPyArg_ParseTuple() *ERROR*\n"));
        iReturn = PYMIDAS_ERR_WRAP_ARG;
    }//if

    DBG((FL,"\t ParamName='%s', ParamName='%s'\n", pszParamName,pszParamValue)); 

    iReturn = SetParam(pszParamName, pszParamValue );

finito:
    DBG((FL,"<-- wrap_SetParam() - end (0x%08X)\n", iReturn));
    return Py_BuildValue("i", iReturn);
}//wrap_SetParam

//------------------------------------------------------------------------
// SetParam()
//
// Set a parameter value.
//
// in       pszParamName	Name of the parameter ("ALL" = show all)
// out		pszParamValue	Value of the parameter
// returns  int        		Error code
//          0           	No error.
//
int ShowParam(char *pszParamName, char **pszParamValue)
{
    int iReturn = 0;
	bool boShowAll = false;
	static char szBuf[1024];
	char szTemp[100];
	
	
    DBG((FL,"==> ShowParam() - begin\n"));
    DBG((FL,"\t ParamName='%s'\n",pszParamName)); 
    
	if ( strcmp(pszParamName, "ALL") == 0 )
		boShowAll = true;
    
    szBuf[0] = '\0';
    
	for ( int i = 0; aParams[i].szName[0] != '\0'; i++ )
	{
		if ( boShowAll || strcmp(aParams[i].szName, pszParamName) == 0 )
		{
			sprintf(szTemp, "%-20.20s %-20.20s %s\n", 
				aParams[i].szName, aParams[i].szValue, 
				(aParams[i].iType==0 ? "STRING":"INTEGER") );

	    	strcat(szBuf, szTemp);
		}//if
	}//for 

	*pszParamValue = szBuf;
	
    DBG((FL,"<-- ShowParam() - end (0x%08X)\n", iReturn));
    return iReturn;
}//SetParam

//--------------------------------------
//--- wrap_ShowParam() Python wrapper ---
//
// ret = PyMidas.ShowParam '<param_name>'
// ret = PyMidas.ShowParam 'ALL'
//
static PyObject *wrap_ShowParam(PyObject *self, PyObject *args)
{
    char *pszParamName  = NULL;
    char *pszParamValue = NULL;
    int iReturn = 0;

    DBG((FL,"================================================================\n"));
    DBG((FL,"==> wrap_ShowParam() - begin\n"));

    if ( !PyArg_ParseTuple(args, "s", &pszParamName) )
    {
        DBG((FL,"\tPyArg_ParseTuple() *ERROR*\n"));
        iReturn = PYMIDAS_ERR_WRAP_ARG;
    }//if

    DBG((FL,"\t ParamName='%s'\n", pszParamName)); 

    iReturn = ShowParam(pszParamName, &pszParamValue );

finito:
    DBG((FL,"<-- wrap_SetParam() - end (0x%08X)\n", iReturn));
    return Py_BuildValue("is", iReturn, pszParamValue );
}//wrap_SetParam

//------------------------------------------------------------------------
// GetMidasStatus()
//
// Get a MIDAS status
//
//
int GetMidasStatus( int *piCmdStat, char **pszResp )
{
    int iReturn = 0;
    int retstat[2];
    static char *pszRespString;
    int iParCommMode = ReadParamInt(PAR_COMM_MODE);
    	
    DBG((FL,"==> GetMidasStatus() - begin\n"));
    
 	iReturn = XCCRDX( giCid, retstat );
    
    DBG((FL,"\t XCCRDX(%d,...)=%d restat=0x%08X 0x%08X\n", 
           giCid, iReturn, retstat[0], retstat[1]));
 
    if ( iReturn == 0 ) 
    {
        // MIDAS is ready and the retstat holds status from command executed 
        // in background Midas.
        DBG((FL,"\t MIDAS_STATUS_READY retstat[0]=0x%08X retstat[1]=0x%08X\n", 
        			retstat[0], retstat[1] ));
        *piCmdStat = retstat[0];
        //
        // We need to get the response file (if any)
        //
        //if ( retstat[0] == MIDAS_RETSTAT_CMDDBCHANGED )
        //{
        //    iReturn = MIDAS_RETSTAT_CMDDBCHANGED;
        //}
        //else 
        if ( retstat[0] == MIDAS_RETSTAT_PARAMREQ )
        {
            // MIDAS has made a request for user to get a missing parameter
            // Used only when on socket mode
            //
            if ( iParCommMode == PARAM_COMM_MODE_SOCKET ) 
            { 
                int iActvals = 0;
                static char szBuffer[80];
                int iUnit = 0;
                int iNull = 0;
                int iKstat = 0;
                
                pszRespString = szBuffer;

                iReturn = XCKRDC( 
                    giCid,      //IN: connection id of background Midas
                    "Q3",       //IN: keyword name (max. 8 chars.)
                    1,          //IN: no. of bytes per data value
                    1,          //IN: first data item to be read
                    60,         //IN: no. of elements to get
                    &iActvals,  //OUT: actual no. of elements returned
                    szBuffer,   //OUT: buffer for data values
                    &iUnit,     //OUT: address of unit-pointer
                    &iNull,     //OUT: no. of null values in keyword
                    &iKstat     //OUT: status from SCKRDC in background Midas
                    );
                if ( iReturn != 0 )
                {
                    // Problems in XCKRDC() function
                    DBG((FL,"\tXCKRDC(..,\"Q3\",..)=%d\n",iReturn)); 
                    iReturn = PYMIDAS_ERR_GENERAL;
                }
                else if ( iKstat != 0 )
                {
                    // Background MIDAS problem
                    DBG((FL,"\tXCKRDC(..,\"Q3\",..): kstat=%d\n",iKstat)); 
                    iReturn = PYMIDAS_ERR_BACKGROUND;
                }
                else
                {
                     DBG((FL,"\tXCKRDC(..,\"Q3\",..) OK a=%d --> '%s'\n",
                            iActvals, szBuffer)); 
                     iReturn = PYMIDAS_ERR_NOERROR_PARAMREQ;
                }//else
                
            }
        }
        else
        {
            // Normal response file
            //
            long lReadCount = ReadResponse(&pszRespString, &iReturn);    
            DBG((FL,"\t lReadCount=%ld \n",lReadCount)); 
    		//DBG((FL,"\t lReadCount=%ld PYMIDAS_ERR_NORESPONSE=%ld\n",lReadCount,PYMIDAS_ERR_NORESPONSE)); 
    		//DBG((FL,"\t lReadCount=%X PYMIDAS_ERR_NORESPONSE=%X\n",lReadCount,PYMIDAS_ERR_NORESPONSE)); 
    
    		if ( lReadCount == 0L ) //PYMIDAS_ERR_NORESPONSE )
    		{
    		    //DBG((FL,"\t PYMIDAS_ERR_NORESPONSE (%X)\n",pszRespString)); 
    			*pszRespString = '\0';
    		    DBG((FL,"\t PYMIDAS_ERR_NORESPONSE (%X)\n",*pszRespString)); 
    		}//if
            else
            {
                DBG((FL,"\tResp='%500.500s'\n",pszRespString)); 
                
            }//else
        }
        *pszResp = pszRespString;
    }//if
   	else
   	{
   		if ( iReturn == -1 ) // problems with socket/file communication 
   			iReturn = PYMIDAS_ERR_NORESPONSE;
   		else
   		if ( iReturn == 1 )
   			iReturn = MIDAS_STATUS_BUSY;
   		else
   		if ( iReturn == 2 )
   			iReturn = MIDAS_STATUS_TIMEOUT;
   		else  //  invalid input to this interface or something else
   			iReturn = PYMIDAS_ERR_GENERAL;
   	}//else
	
    DBG((FL,"<-- GetMidasStatus() - end (0x%08X/%d)\n", iReturn, iReturn));
    return iReturn;
}//GetMidasStatus

//------------------------------------------
//--- wrap_GetMidasStatus Python wrapper ---
//
// ret = PyMidas.ShowParam '<param_name>'
// ret = PyMidas.ShowParam 'ALL'
//
static PyObject *wrap_GetMidasStatus(PyObject *self, PyObject *args)
{
    int iReturn = 0;
    int iCmdStat = 0;
    char *pszResp = NULL;

    DBG((FL,"================================================================\n"));
    DBG((FL,"==> wrap_GetMidasStatus - begin\n"));

    iReturn = GetMidasStatus(&iCmdStat, &pszResp); 

finito:
   // DBG((FL,"\t CmdStat=%d, Resp=%s.\n", iCmdStat, pszResp));
    DBG((FL,"<-- wrap_GetMidasStatus - end (0x%08X)\n", iReturn));
    return Py_BuildValue("iis", iReturn, iCmdStat, pszResp );
}//wrap_GetMidasStatus
//------------------------------------------

//--- wrap_Build Python wrapper ---
//
//
static PyObject *wrap_build(PyObject *self, PyObject *args)
{
    int iReturn = 0;
	char szBuf[50] = "";
	
    DBG((FL,"================================================================\n"));
    DBG((FL,"==> wrap_build - begin\n"));

	strcat(szBuf, __DATE__);
	strcat(szBuf, " ");
	strcat(szBuf, __TIME__);
	
finito:
    DBG((FL,"<-- wrap_build - end (%s)\n", szBuf));
    return Py_BuildValue("s", szBuf );
}//wrap_GetMidasStatus

//------------------------------------------------------------------------
// Exported Python functions.
//
static PyMethodDef PyMidasMethods[] = 
{
	  {"Build", wrap_build, METH_VARARGS, "Build number"},
	  {"Initialize", wrap_Initialize, METH_VARARGS, ""},
	  {"Function", wrap_function, METH_VARARGS, ""},
	  {"Connection", wrap_PymidasConnection, METH_VARARGS, ""},
	  {"GetKeywordValue",wrap_PymidasGetKeywordValue,METH_VARARGS,"Interface!" },
	  {"Do", wrap_PymidasDo, METH_VARARGS, "as a 'drs'"},
	  {"SetParam", wrap_SetParam, METH_VARARGS, "Parameter setup"},
	  {"ShowParam", wrap_ShowParam, METH_VARARGS, "Show all parameter settings"},
	  {"GetMidasStatus", wrap_GetMidasStatus, METH_VARARGS, "Get's MIDAS status"},
	  {NULL, NULL, 0, NULL}        // Sentinel
};
//------------------------------------------------------------------------
// initpymidas()
PyMODINIT_FUNC init_pymidas_c_00(void)
{
    (void) Py_InitModule(PYMIDAS_MODULE_NAME, PyMidasMethods);
    
    //
    // Init some global variables
    // 
    glBufSize = 1024L;
    gpResponseBuf = new char[ glBufSize ];
}//initpymidas()

//------------------------------------------------------------------------
// LayerProlog()
//
// MIDAS Prolog
//
void LayerProlog()
{
	
    if ( giMidasPrologDone )
        return;
    //
    // MIDAS Prolog: should be called once in the main routine before any 
    // other references to the Midas environment are made.
    //
    SCSPRO(PYMIDAS_LAYER_NAME);         
    giMidasPrologDone = 1;
    
    //DBG((FL,"----------------------------------------------------------------\n"));    
    DBG((FL,"- (Build:%s %s) MIDAS Prolog done: %s \n",
        __DATE__,__TIME__, PYMIDAS_LAYER_NAME));
    //DBG((FL,"----------------------------------------------------------------\n"));    
}//LayerProlog()

/*
//------------------------------------------------------------------------
// LayerEpilog()
//
// MIDAS Epilog
// 
void LayerEpilog()
{
    if ( giMidasEpilogDone )
        return;
    //
    // MIDAS Prolog: should be called once in the main routine before any 
    // other references to the Midas environment are made.
    //
    SCSPRO(PYMIDAS_LAYER_NAME);         
    giMidasPrologDone = 1;
}//LayerEpilog()
*/

//------------------------------------------------------------------------
// LOG HANDLING
//
#if PYMIDAS_LOGLEVEL

//------------------------------------------------------------------------
void DebugFileOut(char *pszFile, int iLine, char *pszBuf)
{
    printf("%s",pszBuf);   // output to stdout
    
    FILE *f=fopen(PYMIDAS_LOG_FILE, "at");
    if (f)
    {
        fprintf(f,"%-18.18s %4d: %s", pszFile, iLine,pszBuf);
        fclose(f);
    }//if
}//DebugFileOut

//------------------------------------------------------------------------
//
void DebugFile( char *pszFile, int iLine, char *pszFormat, ... )
{                         // 012345678901234567890
    static char buf[1280] = "yymmdd hhmmss: ";
    va_list ap;
    time_t *tim;
    //struct tm tim;
    long now;
    char szFileName[200];
    char szNow[50];
    int iParTraceLevel = ReadParamInt(PAR_TRACE_LEVEL);
    
    if ( iParTraceLevel == 0 ) // No trace at all
        return;

    int iParTraceConsole = ReadParamInt(PAR_TRACE_CONSOLE);
    int iParTraceFile = ReadParamInt(PAR_TRACE_FILE);
    	
    time(&now);
    strcpy(szNow, ctime(&now));
    szNow[strlen(szNow)-1] = '\0';    // remove '\n'
    
    va_start(ap, pszFormat); 

    vsprintf (&buf[0], pszFormat, ap);

    if ( iParTraceConsole ) 	// At least to the console
	    printf("%s",buf);   	// output to stdout

    if ( iParTraceFile ) 		// To file
	{
		strcpy(szFileName,FRONT.STARTUP);
		//printf("\n**** filename=%s.\n",szFileName);		
		strcat(szFileName,PYMIDAS_LOG_FILE);
		//printf("\n**** filename=%s.\n",szFileName);		
	    FILE *f=fopen(szFileName, "at");
	    if (f)
	    {
	        fprintf(f,"%s %-18.18s %4d: %s", szNow, pszFile, iLine ,buf);
	        fclose(f);
	    }//if
	    else
	        printf("File is not open: %s\n", PYMIDAS_LOG_FILE); 
	}//if
    va_end(ap);
}//DebugFile

#endif  //PYMIDAS_LOGLEVEL

//------------------------------------------------------------------------
// qinit_here()
//
// *** NOT USED ***
//
void qinit_here()
{
int  stat, jj;

char   wstr[160];
char  *cp; 
char  cxp[4];
int   ip[6],fid;
register int  nr;

    DBG((FL,"==> qinit_here() - begin\n"));

    for (jj=0; jj<4; jj++) cxp[jj] = 'N';
    
    /*  get symbols DAZUNIT, MID_WORK (as was done in INPREPA before)  */
    
    OSY_GETSYMB("DAZUNIT",wstr,4);
    gszMidasUnit[0] = FRONT.DAZUNIT[0] = wstr[0];
    gszMidasUnit[1] = FRONT.DAZUNIT[1] = wstr[1];
	gszMidasUnit[2] = '\0';

	gszMyUnit[0] = '0';
	gszMyUnit[1] = '0';
	gszMyUnit[2] = '\0';

    DBG((FL,"\tFRONT.DAZUNIT[0]=%c.\n",FRONT.DAZUNIT[0]));
    DBG((FL,"\tFRONT.DAZUNIT[1]=%c.\n",FRONT.DAZUNIT[1]));
    
    stat = OSY_TRNLOG("MID_WORK",wstr,160,&jj);    /* Decode startup directory */
    DBG((FL,"\tstat=%d jj=%d\n",stat,jj));
    if ( (stat != 0) || (jj > 160) )
    {
	    DBG((FL,"\tERROR!\n"));
       (void) printf
       ("we could not translate MID_WORK or MID_WORK > 160 char.\n\r");
       ospexit(1);
    }
    
    #if vms
    if (wstr[jj-1] != FSY_DISKEND)
       {
       if (wstr[jj-1] != FSY_DIREND)
          {
          wstr[jj++] = FSY_DIREND;
          wstr[jj] = '\0';
          }
       }
    #else
    if (wstr[jj-1] != FSY_DIREND)
       {
       wstr[jj++] = FSY_DIREND;
       wstr[jj] = '\0';
       }
    #endif
    (void) strcpy(FRONT.STARTUP,wstr);  /* save name of startup directory */
    DBG((FL,"\t wstr=[%s]\n",wstr));
    
    (void) strcpy(FRONT.PROMPT,"Midas ");   /* init prompt string  */
    FRONT.PEND = 6;                         /* length of prompt + 1 */

/*
    //  get PID + also store PID in file MIDASxy.PID  
    
    FRONT.PID = oshpid();           //  get process id 
    
    
    jj = CGN_COPY(wstr,FRONT.STARTUP);
    (void) strcpy(&wstr[jj],"MIDAS  .PID");
    wstr[jj+5] = FRONT.DAZUNIT[0];
    wstr[jj+6] = FRONT.DAZUNIT[1];
    fid = osaopen(wstr,1);
    if (fid > 0)
       {
    #if vms
       (void)sprintf(wstr,"%X",FRONT.PID);        // we need PID in hexa ... 
    #else
       (void)sprintf(wstr,"%d",FRONT.PID);
     printf("PID=%d \n",FRONT.PID);
    #endif
       osawrite(fid,wstr,(int)strlen(wstr));
       osaclose(fid);
       }
    else
       (void) printf("Could not build PID file MIDAS%c%c.PID\n\r",
              FRONT.DAZUNIT[0],FRONT.DAZUNIT[1]);
    FRONT.PP = -1;
    FRONT.PLAYBACK = 0;
    FRONT.INTERM = 0;
    FRONT.ENV = '>';
    
    server.MODE = 0;
    server.ECHO = 'N';
    
    for (nr=0; nr<10; nr++)
       {
       OLDTOKEN[nr].STR[0] = '?';
       OLDTOKEN[nr].STR[1] = '\0';
       OLDTOKEN[nr].LEN = 1;
       }

    (void) inxcon(FRONT.DAZUNIT,FRONT.STARTUP);   // init communication stuff
    
    stat= prepx(-1,cxp,ip);         // init on server side (get keys in)
    
    
    
    stat= prepx(-3,cxp,ip);            // do the initial @ login 
*/
#if 0
    int  stat, jj;
    char  wstr[160];

    DBG((FL,"==> qinit_here() - begin\n"));

    FRONT.DAZUNIT[0] = '6';         /* use unit 66 */
    FRONT.DAZUNIT[1] = '6';
 
    stat = OSY_TRNLOG("MID_WORK",wstr,160,&jj);    /* Decode startup directory */
    //DBG((FL,"\tstat=%d wstr='%s' jj=%d\n",stat,wstr,jj));
    
    if (stat != 0) 
    {
        //DBG((FL,"\t\t1\n"));
        stat = OSY_TRNLOG("HOME",wstr,160,&jj);  /* use $HOME/midwork/ */
        if (wstr[jj-1] != FSY_DIREND) wstr[jj++] = FSY_DIREND;
        (void) strcpy(&wstr[jj],"midwork/");  
    } 
    else if (wstr[jj-1] != FSY_DIREND)
    {
        //DBG((FL,"\t\t2\n"));
        wstr[jj++] = FSY_DIREND;
        wstr[jj] = '\0';
    }

    (void) strcpy(FRONT.STARTUP,wstr);  /* save name of startup directory */
    DBG((FL,"\tFRONT.STARTUP='%s' \n",FRONT.STARTUP));

    //FRONT.ENV = '>';        /* to indicate Unix command line input */
    //FRONT.ENV = '<';        /* to indicate Unix command line input */
     FRONT.ENV = '*';        /* to indicate Unix command line input */
    //FRONT.ENV = ':';        /* to indicate Unix command line input */

    //FRONT.PEND = 0;         /* length of prompt + 1 */
    FRONT.PEND = 6;         /* length of prompt + 1 */

    FRONT.PID = 0;          /* not used */
    FRONT.PP = -1;
    FRONT.PLAYBACK = 0;
    FRONT.INTERM = 0;

    //server.MODE = 0;
    server.MODE = 1;
    server.ECHO = 'N';
#endif
    DBG((FL,"<-- qinit_here() - end\n"));
}//qinit_here()

//------------------------------------------------------------------------
#ifdef DBG
#define DBG(arg)
#endif

//------------------------------------------------------------------------
// Tokenizer()
//
char ** Tokenizer( char *pszLine )
{ 
    static char szTokens[MAX_TOKENS][MAX_TOKEN_LEN];
    char *pLine = pszLine;
    int  iToken = 0;
    int  i = 0;

    DBG((FL,"==> Tokenizer() - begin\n"));
    
    if ( pszLine == NULL )
    {
        DBG((FL,"\tNo Line!\n"));
        return NULL;
    }//if
        
    DBG((FL,"\tLine='%s'\n", pszLine));
    
    for (iToken = 0; *pLine != '\0' && iToken < MAX_TOKENS; iToken++)
    {
        //for ( i = 0; *pLine != '\0' && *pLine != ' ' && *pLine != '/'; i++, pLine++ )
        for ( i = 0; *pLine != '\0' && *pLine != ' '; i++, pLine++ )
            szTokens[iToken][i] = *pLine;
            
        szTokens[iToken][i] = '\0';   
        DBG((FL,"\t\tTokens[%d]='%s'\n", iToken, szTokens[iToken]));
        
        if ( *pLine != '\0')
            pLine++;
    }//for
        
    DBG((FL,"<-- Tokenizer() - end\n"));
    return ((char**)szTokens);
}//Tokenizer()

