//########################################################################
//
//   This file is part of PyMidas.
//
//   PyMidas is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   PyMidas is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with PyMidas; if not, write to the Free Software
//   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//
//###########################################################################
/*------------------------------------------------
    pymidas-layer1.h

.VERSION    20050309:  creation  (T.Oittinen)
---------------------------------------------------*/

#ifndef _PYMIDAS1LAYER1_H_
#define _PYMIDAS1LAYER1_H_

//------------------------------------------------------------------------
// Debug log file
// 0 - No debug log file
// 1 - Fatal errors
// 2 - All errors
// 3 - All infos
//
#define PYMIDAS_LOGLEVEL 1

//#define PYMIDAS_LOG_FILE        "/pymidas/pymidas-layer1.log"
//#define PYMIDAS_LOG_FILE        "/home/toittine/pymidas/pymidas-layer1.log"
#define PYMIDAS_LOG_FILE        "pymidas-layer1.log"

#if PYMIDAS_LOGLEVEL

#define FL __FILE__,__LINE__

#ifdef __cplusplus
extern "C" {
#endif

//#define DBG(arg) DebugFile(__FILE__,__LINE__,arg ## )
#define DBG(arg) DebugFile arg
//#define DBG(arg) DebugFile(__FILE__,__LINE__,arg)
//#define DBGMEM(arg) DebugMem arg
//#define F(arg)  DebugFileOut(__FILE__,__LINE__,arg)

//char * DebugFile( char *pszFormat, ... );
//void   DebugFileOut(char *pszFile, int iLine, char *pszBuf);


    void DebugFile( char *pszFile, int iLine, char *pszFormat, ... );
    //void DebugFile( char *pszFormat, ... );
    //void DebugMem( char *pNote, char *pBuf, int iLen );
#ifdef __cplusplus
}
#endif

#else

#define DBG(arg)
#define DBGMEM(arg)

#endif  // PYMIDAS_LOGLEVEL
              
#if OSTYPE_SOLARIS
typedef void (*sighandler_t)(int);
#endif //OSTYPE_SOLARIS
   
//------------------------------------------------------------------------
// Error & return codes
//
#define PYMIDAS_ERR_NOERROR_PARAMREQ    0x00000038
#define PYMIDAS_ERR_NOERROR             0x00000000
#define PYMIDAS_ERR_NOT_SUPPORTED       0xF0010001
#define PYMIDAS_ERR_GENERAL             0xF0010002
#define PYMIDAS_ERR_KEYW_NOT_FOUND      0xF0010004
#define PYMIDAS_ERR_KEYW_VALUE          0xF0010008
#define PYMIDAS_ERR_NORESPONSE          0xF0010010
#define PYMIDAS_ERR_BACKGROUND          0xF0010011

#define PYMIDAS_RET_OK                  PYMIDAS_ERR_NOERROR
#define PYMIDAS_RET_CMDDEFINED          99  // Cmd db of MIDAS has been defined
#define PYMIDAS_RET_CMDDELETED          98  // Cmd db of MIDAS has been deleted

//
//Python wrapper
//
#define PYMIDAS_ERR_WRAP_ARG            0xF0090001
//------------------------------------------------------------------------
// MIDAS Keyword Types
//
#define PYMIDAS_KEYWORD_BLANK           ' '
#define PYMIDAS_KEYWORD_INTEGER         'I'
#define PYMIDAS_KEYWORD_REAL            'R'
#define PYMIDAS_KEYWORD_DOUBLE          'D'
#define PYMIDAS_KEYWORD_CHAR            'C'

//------------------------------------------------------------------------
// MIDAS Status
//
#define MIDAS_STATUS_READY				0	// Cmd completed
#define MIDAS_STATUS_BUSY				1	// Background Midas still busy 
#define MIDAS_STATUS_TIMEOUT			2	// Time out, `secs' seconds have 
											// passed without return 

#define MIDAS_RETSTAT_PARAMREQ          56  // Need a missing parameter
//------------------------------------------------------------------------
// Other
//
#define PYMIDAS_LAYER_NAME              "PyMidas-Layer1"
#define PYMIDAS_MODULE_NAME				"_pymidas_c_00"

#define FILENAME_FORMAT                 "pymidas%2.2s%1d.%04d"
#define INCREQ(x)                       (((x)<9999) ? (x)++ : (x)=0)
//------------------------------------------------------------------------
// Parameters
//
#define PARAM_NAME_LEN 				 31
#define PARAM_VALUE_LEN 			121
#define PARAM_TYPE_STRING   		  0
#define PARAM_TYPE_INT      		  1
#define PARAM_COMM_MODE_FILE      	  0     // use files
#define PARAM_COMM_MODE_SOCKET     	  1     // use sockets
#define PARAM_COMM_OUT_FILE           0     // ">file" 
#define PARAM_COMM_OUT_FILETERM       1     // ">file+terminal"

typedef struct _ParamType
{
	char szName [PARAM_NAME_LEN];
	char szValue[PARAM_VALUE_LEN];
	char szDefValue[PARAM_VALUE_LEN];
	int  iType;			// 0=string  1=int
} ParamType;

#define PAR_TRACE_LEVEL 		"TRACE_LEVEL"
#define PAR_TRACE_CONSOLE		"TRACE_CONSOLE"
#define PAR_TRACE_FILE			"TRACE_FILE"
#define PAR_MIDAS_CMD_TIMEOUT	"MIDAS_CMD_TIMEOUT"
#define PAR_COMM_MODE			"COMM_MODE"
#define PAR_COMM_OUT            "COMM_OUT"

//------------------------------------------------------------------------
// Prototypes for externally used fuctions
//
#ifdef __cplusplus
extern "C" {
#endif
//------------------------------------------------------------------------
// PymidasGetKeywordValue()
//
// Get the value of the given keyword 
// Note: Logfile into PYMIDAS_LOG_FILE !
//
// in       pszKeyName  Name of the keyword
// out      pValue      Value of the requested keyword 
// out      pcType      Type of the keyword
// out      piActLen    Actual length of the value: 
//                      pcType=='C' - length of the string
//                      other - # of elements in an array
// returns  int         Error code:
//          0           No error
int  PymidasGetKeywordValue(
    char *pszKeyName, 
    void *pValue, 
    char *pcType, 
    int *piActLen);
#ifdef __cplusplus
}
#endif

char ** Tokenizer( char *pszLine );


#endif //_PYMIDAS1LAYER1_H_
